from typing import Any, List, Optional

from rich import print
from rich.console import Console
from rich.table import Table


def print_error(msg):
    print(f"[red]{msg}[/red]")


def print_success(msg):
    print(f"[green]{msg}[/green]")


def print_table(headers: List[str], rows: List[List[Any]], style: Optional[str] = None):
    console = Console()
    table = Table(*headers, style=style)
    for row in rows:
        table.add_row(*[str(c) for c in row])
    console.print(table)



class ValidatorError(Exception):
    def __init__(self, msg=None):
        if msg:
            self.msg = msg

    def __str__(self):
        return self.msg or ""


class OpenApiValidationError(ValidatorError):
    pass


class InvalidJSON(OpenApiValidationError):
    msg = "The file does not contain valid JSON"


class UnsupportedVersion(OpenApiValidationError):
    msg = "Validator supports only OpenAPI 3.x specs"


class MissingParameter(OpenApiValidationError):
    pass


class DefinitionError(OpenApiValidationError):
    pass


class WrongContentType(DefinitionError):
    msg = "Model description must be in application/json format"


class SchemaMissing(DefinitionError):
    pass


class NoEndpointsDefined(DefinitionError):
    pass


class OnlyOneEndpointAllowed(DefinitionError):
    pass


class PostMethodIsMissing(DefinitionError):
    pass


class OnlyPostMethodAllowed(DefinitionError):
    pass


class RequestBodyMissing(DefinitionError):
    pass


class ResponseBodyMissing(DefinitionError):
    pass


class AuthorizationHeaderMissing(DefinitionError):
    msg = "Authorization header is missing"


class AuthProviderHeaderMissing(DefinitionError):
    msg = "X-Authorization-Provider header is missing"


class ServersShouldNotBeDefined(DefinitionError):
    msg = '"servers" section should not exist in definition'


class SecurityShouldNotBeDefined(DefinitionError):
    msg = '"security" section should not exist in definition'


class HTTPResponseIsMissing(DefinitionError):
    pass



from pydantic import Field

from definition_tooling.converter import (
    CamelCaseModel,
    DataProductDefinition,
    ErrorResponse,
)


class CoffeeBrewingRequest(CamelCaseModel):
    brew: str = Field(
        ...,
        title="Brew",
        description="Kind of drink to brew",
        example="coffee",
    )


class CoffeeBrewingResponse(CamelCaseModel):
    ok: bool = Field(
        ...,
        title="OK",
        example=True,
    )


@ErrorResponse(description="I'm a teapot")
class TeaPotError(CamelCaseModel):
    ok: bool = Field(
        ...,
        title="OK",
        example=False,
    )
    error_message: str = Field(
        ...,
        title="Error message",
        example="I'm a teapot",
    )


DEFINITION = DataProductDefinition(
    request=CoffeeBrewingRequest,
    response=CoffeeBrewingResponse,
    summary="Coffee brewer",
    error_responses={
        418: TeaPotError,
    },
    deprecated=True,
)



import importlib.util
import json
import subprocess
from pathlib import Path
from typing import Dict, List, Optional, Type

from deepdiff import DeepDiff
from fastapi import FastAPI, Header
from pydantic import BaseModel, ValidationError, conint, validator
from rich import print
from stringcase import camelcase

from definition_tooling.api_errors import DATA_PRODUCT_ERRORS


class CamelCaseModel(BaseModel):
    class Config:
        alias_generator = camelcase
        allow_population_by_field_name = True


class ErrorModel(BaseModel):
    """
    Wrapper used for error responses, see ErrorResponse decorator for more details.

    Encapsulates the actual model and a description for the error.
    """

    model: Type[BaseModel]
    description: str


class ErrorResponse:
    """
    Decorator that should be used around any models that define an error to be used in
    DataProductDefinition.error_responses. It will wrap the class in an ErrorModel in
    order to define and store a custom description for the error in addition to the
    actual model.

    Usage:

    @ErrorResponse(description="Not found")
    class NotFoundResponse(CamelCaseModel):
        ...


    DEFINITION = DataProductDefinition(
        ...
        error_responses={
            404: NotFoundResponse,
        }
    )
    """

    def __init__(self, description: str) -> None:
        self.description = description

    def __call__(self, model_cls: Type[BaseModel]) -> ErrorModel:
        return ErrorModel(
            model=model_cls,
            description=self.description,
        )


ERROR_CODE = conint(ge=400, lt=600)


class DataProductDefinition(BaseModel):
    description: Optional[str]
    name: Optional[str]
    request: Type[BaseModel]
    response: Type[BaseModel]
    route_description: Optional[str]
    route_summary: Optional[str]
    summary: str
    requires_authorization: bool = False
    requires_consent: bool = False
    error_responses: Dict[ERROR_CODE, ErrorModel] = {}
    deprecated: bool = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.summary:
            if not self.route_description:
                self.route_description = self.summary
            if not self.description:
                self.description = self.summary
            if not self.route_summary:
                self.route_summary = self.summary

    @validator("error_responses")
    def validate_error_responses(cls, v: Dict[ERROR_CODE, ErrorModel]):
        status_codes = set(v.keys())
        reserved_status_codes = set(DATA_PRODUCT_ERRORS.keys())
        overlapping = status_codes.intersection(reserved_status_codes)
        if overlapping:
            raise ValueError(
                "Can not contain reserved error code(s): "
                f"{', '.join(str(n) for n in overlapping)}"
            )
        return v


def export_openapi_spec(definition: DataProductDefinition) -> dict:
    """
    Given a data product definition, create a FastAPI application and a corresponding
    POST route. Then export its OpenAPI spec
    :param definition: Data product definition
    :return: OpenAPI spec
    """
    app = FastAPI(
        title=definition.summary,
        description=definition.description,
        version="1.0.0",
    )

    if definition.requires_authorization:
        authorization_header_type = str
        authorization_header_default_value = ...
    else:
        authorization_header_type = Optional[str]
        authorization_header_default_value = None

    if definition.requires_consent:
        consent_header_type = str
        consent_header_default_value = ...
        consent_header_description = "Consent token"
    else:
        consent_header_type = Optional[str]
        consent_header_default_value = None
        consent_header_description = "Optional consent token"

    responses = {
        code: {
            "model": error_model.model,
            "description": error_model.description,
        }
        for code, error_model in definition.error_responses.items()
    }
    responses.update(DATA_PRODUCT_ERRORS)

    @app.post(
        f"/{definition.name}",
        summary=definition.route_summary,
        description=definition.route_description,
        response_model=definition.response,
        responses=responses,
        deprecated=definition.deprecated,
    )
    def request(
        params: definition.request,
        x_consent_token: consent_header_type = Header(
            consent_header_default_value,
            description=consent_header_description,
        ),
        authorization: authorization_header_type = Header(
            authorization_header_default_value,
            description='The login token. Value should be "Bearer [token]"',
        ),
        x_authorization_provider: Optional[str] = Header(
            None, description="The bare domain of the system that provided the token."
        ),
    ):
        pass

    openapi = app.openapi()

    for path, data in openapi["paths"].items():
        operation_id = data["post"]["operationId"].removesuffix("_post")
        openapi["paths"][path]["post"]["operationId"] = operation_id

    return openapi


def styled_error(error: str, path: Path) -> str:
    """
    Style error messages to make them clearer and easier to read
    """
    return f"[bold red]{error}[/bold red] in [yellow]{path}[/yellow]:exclamation:"


def convert_data_product_definitions(src: Path, dest: Path) -> bool:
    """
    Browse folder for definitions defined as python files
    and export them to corresponding OpenAPI specs in the output folder
    """

    should_fail_hook = False
    modified_files = []
    for p in src.glob("**/*.py"):
        spec = importlib.util.spec_from_file_location(name=str(p), location=str(p))
        if not spec.loader:
            raise RuntimeError(f"Failed to import {p} module")
        try:
            module = spec.loader.load_module(str(p))
        except ValidationError as e:
            should_fail_hook = True
            print(styled_error("Validation error", p))
            print(e)
            continue

        try:
            definition: DataProductDefinition = getattr(module, "DEFINITION")
        except AttributeError:
            print(styled_error("Error finding DEFINITION variable", p))
            continue

        # Get definition name based on file path
        definition.name = p.relative_to(src).with_suffix("").as_posix()
        if not definition.route_summary:
            definition.route_summary = definition.name

        openapi = export_openapi_spec(definition)

        out_file = (dest / p.relative_to(src)).with_suffix(".json")

        current_spec = {}
        if out_file.exists():
            current_spec = json.loads(out_file.read_text(encoding="utf-8"))

        # Write resulted JSON only if it's changed to satisfy pre-commit hook
        if DeepDiff(current_spec, openapi, ignore_order=True) != {}:
            print(f"Exporting {out_file}")
            out_file.parent.mkdir(parents=True, exist_ok=True)
            out_file.write_text(
                json.dumps(openapi, indent=2, ensure_ascii=False) + "\n",
                encoding="utf-8",
            )
            modified_files.append(out_file)
            # Hook should fail as we modified the file.
            should_fail_hook = True
        else:
            if file_is_untracked(out_file):
                print(f"Untracked {out_file}")
                should_fail_hook = True
            else:
                print(f"Skipping {out_file}")

    # Run hooks on all modified files at once to save overhead from subprocess
    run_pre_commit_hooks_on_files(modified_files)

    return should_fail_hook


def run_pre_commit_hooks_on_files(files: List[Path]) -> None:
    """
    Run pre-commit hooks on files.
    """
    files = [str(file) for file in files]
    subprocess.run(
        [
            "pre-commit",
            "run",
            "--files",
            *files,
        ],
        capture_output=True,
    )


def file_is_untracked(file: Path) -> bool:
    """
    Check if the file is untracked in git.
    """
    completed_process = subprocess.run(
        ["git", "status", "--short", str(file)],
        check=True,
        capture_output=True,
        encoding="utf-8",
    )
    return completed_process.stdout.startswith("??")



"""
Predefined errors that the product gateway or productizers can return.

These errors can not be overridden by the data product definition itself.
"""
from typing import Optional

from pydantic import BaseModel, Field


class BaseApiError(BaseModel):
    __status__: int


class ApiError(BaseApiError):
    type: str = Field(..., title="Error type", description="Error identifier")
    message: str = Field(..., title="Error message", description="Error description")


class Unauthorized(ApiError):
    __status__ = 401


class Forbidden(ApiError):
    __status__ = 403


class NotFound(ApiError):
    __status__ = 404


# Note: 422 is added automatically by FastAPI


class DataSourceNotFound(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 444
    message: str = Field(
        "Data source not found",
        title="Error message",
        description="Error description",
    )


class DataSourceError(ApiError):
    __status__ = 500


class BadGateway(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 502


class ServiceUnavailable(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 503
    message: Optional[str] = Field(
        None, title="Error message", description="Error description"
    )


class GatewayTimeout(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 504
    message: Optional[str] = Field(
        None, title="Error message", description="Error description"
    )


class DoesNotConformToDefinition(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 550
    message: str = "Response from data source does not conform to definition"
    data_source_status_code: int = Field(
        ...,
        title="Data source status code",
        description="HTTP status code returned from the data source",
    )


DATA_PRODUCT_ERRORS = {
    resp.__status__: {"model": resp}
    for resp in [
        Unauthorized,
        Forbidden,
        NotFound,
        DataSourceNotFound,
        DataSourceError,
        BadGateway,
        ServiceUnavailable,
        GatewayTimeout,
        DoesNotConformToDefinition,
    ]
}



from pydantic import Field

from definition_tooling.converter import (
    CamelCaseModel,
    DataProductDefinition,
    ErrorResponse,
)


class BasicCompanyInfoRequest(CamelCaseModel):
    company_id: str = Field(
        ...,
        title="Company ID",
        description="The ID of the company",
        example="2464491-9",
    )


class BasicCompanyInfoResponse(CamelCaseModel):
    name: str = Field(
        ..., title="Name of the company", example="Digital Living International Oy"
    )
    company_id: str = Field(..., title="ID of the company", example="2464491-9")
    company_form: str = Field(
        ..., title="The company form of the company", example="LLC"
    )
    registration_date: str = Field(
        ..., title="Date of registration for the company", example="2012-02-23"
    )


@ErrorResponse(description="Unavailable for some legal reasons")
class UnavailableForLegalReasonsResponse(CamelCaseModel):
    reasons: str = Field(
        ...,
        title="Reason",
        description="The reason why the data is not available",
    )


DEFINITION = DataProductDefinition(
    description="Data Product for basic company info",
    request=BasicCompanyInfoRequest,
    response=BasicCompanyInfoResponse,
    route_description="Information about the company",
    summary="Basic Company Info",
    error_responses={
        422: UnavailableForLegalReasonsResponse,
    },
)



import json
from copy import deepcopy
from pathlib import Path

import pytest

from definition_tooling.validator import errors as err
from definition_tooling.validator.core import DefinitionValidator

# Note: It's easier to get some 100% valid spec and corrupt it
# instead of having multiple incorrect specs in the repo

SPECS_ROOT_DIR = Path(__file__).absolute().parent / "data"
COMPANY_BASIC_INFO: dict = json.loads(
    (SPECS_ROOT_DIR / "CompanyBasicInfo.json").read_text(encoding="utf8")
)


def check_validation_error(tmp_path, spec: dict, exception):
    spec_path = tmp_path / "spec.json"
    spec_path.write_text(json.dumps(spec))
    with pytest.raises(exception):
        DefinitionValidator(spec_path).validate()


@pytest.mark.parametrize("method", ["get", "put", "delete"])
def test_standards_has_non_post_method(method, tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["paths"]["/draft/Company/BasicInfo"][method] = {
        "description": "Method which should not exist"
    }
    check_validation_error(tmp_path, spec, err.OnlyPostMethodAllowed)


def test_post_method_is_missing(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["paths"]["/draft/Company/BasicInfo"]["post"]
    check_validation_error(tmp_path, spec, err.PostMethodIsMissing)


def test_many_endpoints(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["paths"]["/pets"] = {"post": {"description": "Pet store, why not?"}}
    check_validation_error(tmp_path, spec, err.OnlyOneEndpointAllowed)


def test_no_endpoints(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["paths"]
    check_validation_error(tmp_path, spec, err.NoEndpointsDefined)


def test_missing_field_body_is_fine(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["paths"]["/draft/Company/BasicInfo"]["post"]["requestBody"]
    spec_path = tmp_path / "spec.json"
    spec_path.write_text(json.dumps(spec))
    DefinitionValidator(spec_path).validate()


def test_missing_200_response(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["paths"]["/draft/Company/BasicInfo"]["post"]["responses"]["200"]
    check_validation_error(tmp_path, spec, err.ResponseBodyMissing)


def test_wrong_content_type_of_request_body(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    request_body = spec["paths"]["/draft/Company/BasicInfo"]["post"]["requestBody"]
    schema = deepcopy(request_body["content"]["application/json"])
    request_body["content"]["text/plan"] = schema
    del request_body["content"]["application/json"]
    check_validation_error(tmp_path, spec, err.WrongContentType)


def test_wrong_content_type_of_response(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    response = spec["paths"]["/draft/Company/BasicInfo"]["post"]["responses"]["200"]
    schema = deepcopy(response["content"]["application/json"])
    response["content"]["text/plan"] = schema
    del response["content"]["application/json"]
    check_validation_error(tmp_path, spec, err.WrongContentType)


def test_component_schema_is_missing(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["components"]["schemas"]
    check_validation_error(tmp_path, spec, err.SchemaMissing)


@pytest.mark.parametrize(
    "model_name", ["BasicCompanyInfoRequest", "BasicCompanyInfoResponse"]
)
def test_component_is_missing(model_name, tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    del spec["components"]["schemas"][model_name]
    check_validation_error(tmp_path, spec, err.SchemaMissing)


def test_non_existing_component_defined_in_body(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    body = spec["paths"]["/draft/Company/BasicInfo"]["post"]["requestBody"]
    body["content"]["application/json"]["schema"]["$ref"] += "blah"
    check_validation_error(tmp_path, spec, err.SchemaMissing)


def test_non_existing_component_defined_in_response(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    resp_200 = spec["paths"]["/draft/Company/BasicInfo"]["post"]["responses"]["200"]
    resp_200["content"]["application/json"]["schema"]["$ref"] += "blah"
    check_validation_error(tmp_path, spec, err.SchemaMissing)


def test_auth_header_is_missing(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    x_app_provider_header = {
        "schema": {"type": "string"},
        "in": "header",
        "name": "X-Authorization-Provider",
        "description": "Provider domain",
    }
    spec["paths"]["/draft/Company/BasicInfo"]["post"]["parameters"] = [
        x_app_provider_header
    ]
    check_validation_error(tmp_path, spec, err.AuthorizationHeaderMissing)


def test_auth_provider_header_is_missing(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    auth_header = {
        "schema": {"type": "string"},
        "in": "header",
        "name": "Authorization",
        "description": "User bearer token",
    }
    spec["paths"]["/draft/Company/BasicInfo"]["post"]["parameters"] = [auth_header]
    check_validation_error(tmp_path, spec, err.AuthProviderHeaderMissing)


def test_servers_are_defined(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["servers"] = [{"url": "http://example.com"}]
    check_validation_error(tmp_path, spec, err.ServersShouldNotBeDefined)


def test_security_is_defined(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["paths"]["/draft/Company/BasicInfo"]["post"]["security"] = {}
    check_validation_error(tmp_path, spec, err.SecurityShouldNotBeDefined)


def test_loading_non_json_file(tmp_path):
    spec_path = tmp_path / "spec.json"
    spec_path.write_text("weirdo content")
    with pytest.raises(err.InvalidJSON):
        DefinitionValidator(spec_path).validate()


def test_loading_unsupported_version(tmp_path):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["openapi"] = "999.999.999"
    check_validation_error(tmp_path, spec, err.UnsupportedVersion)


@pytest.mark.parametrize("code", [401, 403, 404, 422, 444, 502, 503, 504, 550])
def test_http_errors_defined(tmp_path, code):
    spec = deepcopy(COMPANY_BASIC_INFO)
    spec["paths"]["/draft/Company/BasicInfo"]["post"]["responses"].pop(str(code), None)
    check_validation_error(tmp_path, spec, err.HTTPResponseIsMissing)



import json
from pathlib import Path
from typing import Union

from definition_tooling.api_errors import DATA_PRODUCT_ERRORS
from definition_tooling.validator import errors as err


def validate_component_schema(spec: dict, components_schema: dict):
    if not spec["content"].get("application/json"):
        raise err.WrongContentType
    ref = spec["content"]["application/json"].get("schema", {}).get("$ref")
    if not ref:
        raise err.SchemaMissing(
            'Request or response model is missing from "schema/$ref" section'
        )
    if not ref.startswith("#/components/schemas/"):
        raise err.SchemaMissing(
            "Request and response models must be defined in the"
            '"#/components/schemas/" section'
        )
    model_name = ref.split("/")[-1]
    if not components_schema.get(model_name):
        raise err.SchemaMissing(f"Component schema is missing for {model_name}")


def validate_spec(spec: dict):
    """
    Validate that OpenAPI spec looks like a data product definition. For example, that
    it only has one POST method defined.

    :param spec: OpenAPI spec
    :raises OpenApiValidationError: When OpenAPI spec is incorrect
    """
    if "servers" in spec:
        raise err.ServersShouldNotBeDefined

    if not spec.get("openapi", "").startswith("3"):
        raise err.UnsupportedVersion("Validator supports only OpenAPI 3.x specs")

    paths = spec.get("paths", {})
    if not paths:
        raise err.NoEndpointsDefined
    if len(paths) > 1:
        raise err.OnlyOneEndpointAllowed

    post_route = {}
    for name, path in paths.items():
        methods = list(path)
        if "post" not in methods:
            raise err.PostMethodIsMissing
        if methods != ["post"]:
            raise err.OnlyPostMethodAllowed
        post_route = path["post"]

    component_schemas = spec.get("components", {}).get("schemas")
    if not component_schemas:
        raise err.SchemaMissing('No "components/schemas" section defined')

    if "security" in post_route:
        raise err.SecurityShouldNotBeDefined

    if post_route.get("requestBody", {}).get("content"):
        validate_component_schema(post_route["requestBody"], component_schemas)

    responses = post_route.get("responses", {})
    if not responses.get("200") or not responses["200"].get("content"):
        raise err.ResponseBodyMissing
    validate_component_schema(responses["200"], component_schemas)

    for code in {*DATA_PRODUCT_ERRORS, 422}:
        if not responses.get(str(code)):
            raise err.HTTPResponseIsMissing(f"Missing response for status code {code}")

    headers = [
        param.get("name", "").lower()
        for param in post_route.get("parameters", [])
        if param.get("in") == "header"
    ]
    if "authorization" not in headers:
        raise err.AuthorizationHeaderMissing
    if "x-authorization-provider" not in headers:
        raise err.AuthProviderHeaderMissing


class DefinitionValidator:
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def validate(self):
        try:
            spec = json.loads(self.path.read_text(encoding="utf8"))
        except json.JSONDecodeError:
            raise err.InvalidJSON(f"Incorrect JSON: {self.path}")
        except Exception as e:
            raise err.ValidatorError(f"Failed to validate {self.path}: {e}")
        self.validate_spec(spec)

    @classmethod
    def validate_spec(cls, spec: dict):
        # it's moved to separate function to reduce indentation
        return validate_spec(spec)



import sys
import traceback
from contextlib import contextmanager
from pathlib import Path

from rich import print
from typer import Argument, Typer

from definition_tooling.log import print_error, print_success, print_table
from definition_tooling.validator.core import DefinitionValidator
from definition_tooling.validator.errors import ValidatorError

cli = Typer()


def print_dashes(char="-", length=79):
    print(char * length)


@contextmanager
def header():
    print_dashes(char="=")
    yield
    print_dashes(char="=")


@cli.command(help="Validate Data Product definitions")
def cli_validate_specs(
    path: str = Argument(..., help="Path to directory with definitions in JSON format")
):
    sys.exit(validate_specs(Path(path)))


def validate_specs(path: Path) -> int:
    with header():
        print(f"OpenAPI specs root path: {path}")

    passed, failed = 0, 0
    for spec_path in path.glob("**/*.json"):
        print(f"File: {spec_path}")
        try:
            DefinitionValidator(spec_path).validate()
        except Exception as exc:
            if isinstance(exc, ValidatorError):
                detail = ": " + str(exc) if str(exc) else ""
                print(f"{exc.__class__.__name__}{detail}")
            else:
                print("\n", traceback.format_exc())
            print_error("[FAILED]")
            failed += 1
        else:
            print_success("[PASSED]")
            passed += 1
        print_dashes("-")

    print_table(
        ["Summary", "#"],
        [
            ["Passed", passed],
            ["Failed", failed],
            ["Total", passed + failed],
        ],
        "green" if not failed else "red",
    )

    return 1 if failed else 0
